import subprocess
import attr

from .common import Driver
from ..factory import target_factory
from ..step import step


def _connect_callback(element, pad, target):
    pad.link(target)


@target_factory.reg_driver
@attr.s(eq=False)
class USBAudioInputDriver(Driver):
    """
    This driver provides access to a USB audio input device using ALSA and gstreamer.

    When using this driver in a Python venv, you may need to allow access to
    the gi (GObject Introspection) module from the system. This can be done by
    crating a symlink:
    ln -s /usr/lib/python3/dist-packages/gi $VENV/lib/python*/site-packages/
    """

    bindings = {
        "res": {"USBAudioInput", "NetworkUSBAudioInput"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._prepared = False

    def _get_pipeline(self):
        return ["alsasrc", f"device={self.res.alsa_name}", "!", "matroskamux", "streamable=true", "!", "fdsink"]

    def _prepare(self):
        if self._prepared:
            return

        import gi

        gi.require_version("Gst", "1.0")
        gi.require_version("GLib", "2.0")
        gi.require_version("GObject", "2.0")
        from gi.repository import GLib, GObject, Gst

        self._GLib = GLib
        self._GObject = GObject
        self._Gst = Gst

        Gst.init(None)

        class USBAudioInputBin(Gst.Bin):
            def __init__(self, *, sender, logger):
                super(Gst.Bin, self).__init__()
                self._sender = sender
                self._logger = logger

                assert sender.poll() is None, "sender must be running"
                src = Gst.ElementFactory.make("fdsrc")
                src.set_property("fd", sender.stdout.fileno())
                demux = Gst.ElementFactory.make("matroskademux")
                convert = Gst.ElementFactory.make("audioconvert")

                self.add(src, demux, convert)

                src.link(demux)

                demux.connect("pad-added", _connect_callback, convert.get_static_pad("sink"))

                self.add_pad(Gst.GhostPad.new("src", convert.srcpads[0]))

            def __del__(self):
                self._logger.debug("stopping sender")
                self._sender.terminate()
                try:
                    self._sender.wait(timeout=0.1)
                except subprocess.TimeoutExpired:
                    pass

                self._sender.kill()
                del self._sender

        self._USBAudioInputBin = USBAudioInputBin

    @Driver.check_active
    @step()
    def start_sender(self):
        """Return a subprocess which provides audio data in a matroska container on stdout"""
        tx_cmd = self.res.command_prefix + ["gst-launch-1.0", "-q"]
        tx_cmd += self._get_pipeline()

        tx = subprocess.Popen(
            tx_cmd,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
        )

        return tx

    @Driver.check_active
    @step()
    def create_gst_src(self):
        """Returns a newly create gstreamer bin with a single audio output pad."""
        self._prepare()

        tx = self.start_sender()
        return self._USBAudioInputBin(sender=tx, logger=self.logger)

    @Driver.check_active
    @step(result=True)
    def measure_level(self):
        """Returns the current peak and rms value (measured with the gst level element)"""
        self._prepare()
        Gst = self._Gst
        GLib = self._GLib

        src = self.create_gst_src()

        pipe = Gst.Pipeline.new("dynamic")
        level = Gst.ElementFactory.make("level")
        sink = Gst.ElementFactory.make("fakesink")

        pipe.add(src, level, sink)

        src.link(level)
        level.link(sink)

        loop = GLib.MainLoop()

        stats = {}

        def bus_call(bus, message, loop):
            t = message.type
            if t == Gst.MessageType.EOS:
                self.logger.info("End-of-stream")
                loop.quit()
            elif t == Gst.MessageType.ERROR:
                err, debug = message.parse_error()
                self.logger.error("Error: %s: %s", err, debug)
                loop.quit()
            elif t == Gst.MessageType.ELEMENT:
                self.logger.debug("peak %s", message.get_structure()["peak"])
                self.logger.debug("rms %s", message.get_structure()["rms"])
                stats.update(dict(message.get_structure()))
                loop.quit()
            else:
                self.logger.debug("gst message %s", t)
            return True

        bus = pipe.get_bus()
        bus.add_signal_watch()
        bus.connect("message", bus_call, loop)

        def timeout(*args):
            self.logger.error("loop timed out")
            stats["timeout"] = True
            loop.quit()

        GLib.timeout_add_seconds(15, timeout, None)

        # start play back and listen to events
        pipe.set_state(Gst.State.PLAYING)
        try:
            loop.run()
        finally:
            pipe.set_state(Gst.State.NULL)

        if stats.get("timeout"):
            raise TimeoutError("no data received before timeout")
        return stats

    @Driver.check_active
    @step()
    def play(self):
        """Plays the captured audio via gstreamer's autoaudiosink"""
        tx = self.start_sender()

        rx_cmd = ["gst-launch-1.0"]
        rx_cmd += "fdsrc ! matroskademux ! audioconvert ! autoaudiosink".split()
        rx = subprocess.Popen(
            rx_cmd,
            stdin=tx.stdout,
            stdout=subprocess.DEVNULL,
        )

        # wait until one subprocess has termianted
        while True:
            try:
                tx.wait(timeout=0.1)
                break
            except subprocess.TimeoutExpired:
                pass
            try:
                rx.wait(timeout=0.1)
                break
            except subprocess.TimeoutExpired:
                pass

        rx.terminate()
        tx.terminate()

        rx.communicate()
        tx.communicate()

        return tx.returncode or rx.returncode



import os
import tempfile


def atomic_replace(filename, data):
    f = None
    try:
        with tempfile.NamedTemporaryFile(mode="wb", dir=os.path.dirname(filename), delete=False) as f:
            f.write(data)
            f.flush()
            os.fsync(f.fileno())
        os.replace(f.name, filename)
    finally:
        if f is not None:
            try:
                os.unlink(f.name)
            except FileNotFoundError:
                pass



"""All GPIO-related drivers"""
import attr

from ..factory import target_factory
from ..protocol import DigitalOutputProtocol
from ..resource.remote import NetworkSysfsGPIO
from ..step import step
from .common import Driver
from ..util.agentwrapper import AgentWrapper


@target_factory.reg_driver
@attr.s(eq=False)
class GpioDigitalOutputDriver(Driver, DigitalOutputProtocol):

    bindings = {
        "gpio": {"SysfsGPIO", "MatchedSysfsGPIO", "NetworkSysfsGPIO"},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.wrapper = None

    def on_activate(self):
        if isinstance(self.gpio, NetworkSysfsGPIO):
            host = self.gpio.host
        else:
            host = None
        self.wrapper = AgentWrapper(host)
        self.proxy = self.wrapper.load('sysfsgpio')

    def on_deactivate(self):
        self.wrapper.close()
        self.wrapper = None
        self.proxy = None

    @Driver.check_active
    @step(args=['status'])
    def set(self, status):
        self.proxy.set(self.gpio.index, status)

    @Driver.check_active
    @step(result=True)
    def get(self):
        return self.proxy.get(self.gpio.index)



import os.path
import re
import subprocess
import shutil
import tempfile
import time
import uuid
import csv

from time import sleep

import attr

from ..factory import target_factory
from ..protocol import PowerProtocol
from ..resource.remote import NetworkSigrokUSBDevice, NetworkSigrokUSBSerialDevice
from ..resource.udev import SigrokUSBDevice, SigrokUSBSerialDevice
from ..resource.sigrok import SigrokDevice
from ..step import step
from ..util.helper import processwrapper
from .common import Driver, check_file
from .exception import ExecutionError
from .powerdriver import PowerResetMixin
from ..util import Timeout


@attr.s(eq=False)
class SigrokCommon(Driver):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        # FIXME make sure we always have an environment or config
        if self.target.env:
            self.tool = self.target.env.config.get_tool(
                'sigrok-cli'
            ) or 'sigrok-cli'
        else:
            self.tool = 'sigrok-cli'
        self._running = False

    def _create_tmpdir(self):
        if isinstance(self.sigrok, NetworkSigrokUSBDevice):
            self._tmpdir = f'/tmp/labgrid-sigrok-{uuid.uuid1()}'
            command = self.sigrok.command_prefix + [
                'mkdir', '-p', self._tmpdir
            ]
            self.logger.debug("Tmpdir command: %s", command)
            subprocess.call(
                command,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )
            self.logger.debug("Created tmpdir: %s", self._tmpdir)
            self._local_tmpdir = tempfile.mkdtemp(prefix="labgrid-sigrok-")
            self.logger.debug("Created local tmpdir: %s", self._local_tmpdir)
        else:
            self._tmpdir = tempfile.mkdtemp(prefix="labgrid-sigrok-")
            self.logger.debug("created tmpdir: %s", self._tmpdir)

    def _delete_tmpdir(self):
        if isinstance(self.sigrok, NetworkSigrokUSBDevice):
            command = self.sigrok.command_prefix + [
                'rm', '-r', self._tmpdir
            ]
            self.logger.debug("Tmpdir command: %s", command)
            subprocess.call(
                command,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )
            shutil.rmtree(self._local_tmpdir)
        else:
            shutil.rmtree(self._tmpdir)

    def on_activate(self):
        self._create_tmpdir()

    def on_deactivate(self):
        self._delete_tmpdir()

    def _get_sigrok_prefix(self):
        prefix = [self.tool]
        if isinstance(self.sigrok, (NetworkSigrokUSBDevice, SigrokUSBDevice)):
            prefix += ["-d", f"{self.sigrok.driver}:conn={self.sigrok.busnum}.{self.sigrok.devnum}"]
        elif isinstance(self.sigrok, (NetworkSigrokUSBSerialDevice, SigrokUSBSerialDevice)):
            prefix += ["-d", f"{self.sigrok.driver}:conn={self.sigrok.path}"]
        else:
            prefix += ["-d", self.sigrok.driver]
        if self.sigrok.channels:
            prefix += ["-C", self.sigrok.channels]
        if self.sigrok.channel_group:
            prefix += ["-g", self.sigrok.channel_group]
        return self.sigrok.command_prefix + prefix

    @Driver.check_active
    @step(title='call', args=['args'])
    def _call_with_driver(self, *args):
        combined = self._get_sigrok_prefix() + list(args)
        self.logger.debug("Combined command: %s", " ".join(combined))
        self._process = subprocess.Popen(
            combined,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

    @Driver.check_active
    @step(title='call', args=['args'])
    def _call(self, *args):
        combined = self.sigrok.command_prefix + [self.tool]
        if self.sigrok.channels:
            combined += ["-C", self.sigrok.channels]
        if self.sigrok.channel_group:
            combined += ["-g", self.sigrok.channel_group]
        combined += list(args)
        self.logger.debug("Combined command: %s", combined)
        self._process = subprocess.Popen(
            combined,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE
        )


@target_factory.reg_driver
@attr.s(eq=False)
class SigrokDriver(SigrokCommon):
    """The SigrokDriver uses sigrok-cli to record samples and expose them as python dictionaries.

    Args:
        bindings (dict): driver to use with sigrok
    """
    bindings = {
        "sigrok": {SigrokUSBDevice, NetworkSigrokUSBDevice, SigrokDevice},
    }

    @Driver.check_active
    def capture(self, filename, samplerate="200k"):
        self._filename = filename
        self._basename = os.path.basename(self._filename)
        self.logger.debug(
            "Saving to: %s with basename: %s", self._filename, self._basename
        )
        cmd = [
            "-l", "4", "--config", f"samplerate={samplerate}",
            "--continuous", "-o"
        ]
        filename = os.path.join(self._tmpdir, self._basename)
        cmd.append(filename)
        self._call_with_driver(*cmd)
        args = self.sigrok.command_prefix + ['test', '-e', filename]

        while subprocess.call(args):
            # in case the sigrok-cli call fails, this would wait forever.
            # to avoid this, we also check the spawned sigrok process
            if self._process.poll() is not None:
                ret = self._process.returncode
                if ret != 0:
                    stdout, stderr = self._process.communicate()
                    self.logger.debug("sigrok-cli call terminated prematurely with non-zero return-code")
                    self.logger.debug("stdout: %s", stdout)
                    self.logger.debug("stderr: %s", stderr)
                    raise ExecutionError(f"sigrok-cli call terminated prematurely with return-code '{ret}'.")
            sleep(0.1)

        self._running = True

    @Driver.check_active
    def stop(self):
        assert self._running
        self._running = False
        fnames = ['time']
        fnames.extend(self.sigrok.channels.split(','))
        csv_filename = f'{os.path.splitext(self._basename)[0]}.csv'

        # sigrok-cli can be quit through any keypress
        stdout, stderr = self._process.communicate(input=b"q")
        self.logger.debug("stdout: %s", stdout)
        self.logger.debug("stderr: %s", stderr)

        # Convert from .sr to .csv
        cmd = [
            '-i',
            os.path.join(self._tmpdir, self._basename), '-O', 'csv:time=true', '-o',
            os.path.join(self._tmpdir, csv_filename)
        ]
        self._call(*cmd)
        stdout, stderr = self._process.communicate()
        self.logger.debug("stdout: %s", stdout)
        self.logger.debug("stderr: %s", stderr)
        if isinstance(self.sigrok, NetworkSigrokUSBDevice):
            subprocess.call([
                'scp', f'{self.sigrok.host}:{os.path.join(self._tmpdir, self._basename)}',
                os.path.abspath(self._filename)
            ],
                            stdin=subprocess.DEVNULL,
                            stdout=subprocess.DEVNULL,
                            stderr=subprocess.DEVNULL)
            # get csv from remote host
            subprocess.call([
                'scp', f'{self.sigrok.host}:{os.path.join(self._tmpdir, csv_filename)}',
                os.path.join(self._local_tmpdir, csv_filename)
            ],
                            stdin=subprocess.DEVNULL,
                            stdout=subprocess.DEVNULL,
                            stderr=subprocess.DEVNULL)
            with open(os.path.join(self._local_tmpdir,
                                   csv_filename)) as csv_file:
                # skip first 5 lines of the csv output, contains metadata and fieldnames
                for _ in range(0, 5):
                    next(csv_file)
                return list(csv.DictReader(csv_file, fieldnames=fnames))
        else:
            shutil.copyfile(
                os.path.join(self._tmpdir, self._basename), self._filename
            )
            with open(os.path.join(self._tmpdir, csv_filename)) as csv_file:
                # skip first 5 lines of the csv output, contains metadata and fieldnames
                for _ in range(0, 5):
                    next(csv_file)
                return list(csv.DictReader(csv_file, fieldnames=fnames))

    @Driver.check_active
    def analyze(self, args, filename=None):
        annotation_regex = re.compile(r'(?P<startnum>\d+)-(?P<endnum>\d+) (?P<decoder>[\w\-]+): (?P<annotation>[\w\-]+): (?P<data>".*)')  # pylint: disable=line-too-long
        if not filename and self._filename:
            filename = os.path.join(self._tmpdir, self._basename)
        else:
            filename = os.path.abspath(filename)
        check_file(filename, command_prefix=self.sigrok.command_prefix)
        args.insert(0, filename)
        if isinstance(args, str):
            args = args.split(" ")
        args.insert(0, '-i')
        args.append("--protocol-decoder-samplenum")
        args.append("-l")
        args.append("4")
        combined = self._get_sigrok_prefix() + args
        output = subprocess.check_output(combined)
        return [
            match.groupdict()
            for match in re.finditer(annotation_regex, output.decode("utf-8"))
        ]


@target_factory.reg_driver
@attr.s(eq=False)
class SigrokPowerDriver(SigrokCommon, PowerResetMixin, PowerProtocol):
    """The SigrokPowerDriverDriver uses sigrok-cli to control a PSU and collect
    measurements.

    Args:
        bindings (dict): driver to use with sigrok
    """
    bindings = {
        "sigrok": {SigrokUSBSerialDevice, NetworkSigrokUSBSerialDevice},
    }
    delay = attr.ib(default=3.0, validator=attr.validators.instance_of(float))
    max_voltage = attr.ib(
        default=None,
        converter=attr.converters.optional(float),
        validator=attr.validators.optional(attr.validators.instance_of(float)),
    )
    max_current = attr.ib(
        default=None,
        converter=attr.converters.optional(float),
        validator=attr.validators.optional(attr.validators.instance_of(float)),
    )

    @Driver.check_active
    @step()
    def on(self):
        processwrapper.check_output(
            self._get_sigrok_prefix() + ["--config", "enabled=yes", "--set"]
        )

    @Driver.check_active
    @step()
    def off(self):
        processwrapper.check_output(
            self._get_sigrok_prefix() + ["--config", "enabled=no", "--set"]
        )

    @Driver.check_active
    @step()
    def cycle(self):
        self.off()
        time.sleep(self.delay)
        self.on()

    @Driver.check_active
    @step(args=["value"])
    def set_voltage_target(self, value):
        if self.max_voltage is not None and value > self.max_voltage:
            raise ValueError(
                f"Requested voltage target({value}) is higher than configured maximum ({self.max_voltage})")  # pylint: disable=line-too-long
        processwrapper.check_output(
            self._get_sigrok_prefix() + ["--config", f"voltage_target={value:f}", "--set"]
        )

    @Driver.check_active
    @step(args=["value"])
    def set_current_limit(self, value):
        if self.max_current is not None and value > self.max_current:
            raise ValueError(
                f"Requested current limit ({value}) is higher than configured maximum ({self.max_current})")  # pylint: disable=line-too-long
        processwrapper.check_output(
            self._get_sigrok_prefix() + ["--config", f"current_limit={value:f}", "--set"]
        )

    @Driver.check_active
    @step(result=True)
    def get(self):
        out = processwrapper.check_output(
            self._get_sigrok_prefix() + ["--get", "enabled"]
        )
        if out == b'true':
            return True
        elif out == b'false':
            return False

        raise ExecutionError(f"Unkown enable status {out}")

    @Driver.check_active
    @step(result=True)
    def measure(self):
        out = processwrapper.check_output(
            self._get_sigrok_prefix() + ["--show"]
        )
        res = {}
        for line in out.splitlines():
            line = line.strip()
            if b':' not in line:
                continue
            k, v = line.split(b':', 1)
            if k == b'voltage':
                res['voltage'] = float(v)
            elif k == b'current':
                res['current'] = float(v)
        if len(res) != 2:
            raise ExecutionError(f"Cannot parse --show output {out}")
        return res

@target_factory.reg_driver
@attr.s(eq=False)
class SigrokDmmDriver(SigrokCommon):
    """
    This driver wraps around a single channel DMM controlled by sigrok.
    It has been tested with Unit-T UT61C and UT61B devices but probably also
    works with other single chnnel DMMs.

    This driver binds to a SigrokUsbDevice.
    Make sure to select the correct driver for your DMM there.

    Example usage:
    > resources:
    >   - SigrokUSBDevice:
    >       driver: uni-t-ut61c
    >       match:
    >         'ID_PATH': pci-0000:07:00.4-usb-0:2:1.0
    > drivers:
    >   - SigrokDmmDriver: {}

    Args:
        bindings (dict): driver to use with sigrok
    """

    bindings = {
        "sigrok": {SigrokUSBSerialDevice, NetworkSigrokUSBSerialDevice, SigrokUSBDevice, NetworkSigrokUSBDevice},
    }

    @Driver.check_active
    @step(result=True)
    def capture(self, samples, timeout=None):
        """
        Starts to read samples from the DMM.
        This method returns once sampling has been started. Sampling continues in the background.

        Note: We use subprocess.PIPE to buffer the samples.
        When this buffer is too small for the number of samples requested sampling may stall.

        Args:
            samples: Number of samples to obtain
            timeout: Timeout after which sampling should be stopped.
                     If None: timeout[s] = samples * 1s + 5s
                     If int: Timeout in [s]

        Raises:
            RuntimeError() if a capture is already running.
        """
        if self._running:
            raise RuntimeError("capture is already running")

        if not timeout:
            timeout = samples + 5.0

        args = f"-O csv --samples {samples}".split(" ")
        self._call_with_driver(*args)
        self._timeout = Timeout(timeout)
        self._running = True

    @Driver.check_active
    @step(result=True)
    def stop(self):
        """
        Waits for sigrok to complete and returns all samples obtained afterwards.
        This function blocks until either sigrok has terminated or the timeout has been reached.

        Returns:
            (unit_spec, [sample, ...])

        Raises:
            RuntimeError() if capture has not been started
        """
        if not self._running:
            raise RuntimeError("no capture started yet")
        while not self._timeout.expired:
            if self._process.poll() is not None:
                # process has finished. no need to wait for the timeout
                break
            time.sleep(0.1)
        else:
            # process did not finish in time
            self.logger.info("sigrok-cli did not finish in time, increase timeout?")
            self._process.kill()

        res = []
        unit = ""
        for line in self._process.stdout.readlines():
            line = line.strip()
            if b";" in line:
                # discard header information
                continue
            if not unit:
                # the first line after the header contains the unit information
                unit = line.decode()
            else:
                # all other lines are actual values
                res.append(float(line))
        _, stderr = self._process.communicate()
        self.logger.debug("stderr: %s", stderr)

        self._running = False
        return unit, res

    def on_activate(self):
        # This driver does not use self._tmpdir from SigrockCommon.
        # Overriding this function to inhibit the temp-dir creation.
        pass

    def on_deactivate(self):
        # This driver does not use self._tmpdir from SigrockCommon.
        # Overriding this function to inhibit the temp-dir creation.
        pass



import json
import logging
import attr


@attr.s(eq=False)
class QMPMonitor:
    monitor_out = attr.ib()
    monitor_in = attr.ib()

    def __attrs_post_init__(self):
        self.logger = logging.getLogger(f"{self}")
        self._negotiate_capabilities()

    def _negotiate_capabilities(self):
        greeting = self._read_parse_json()
        if not greeting.get("QMP"):
            raise QMPError("QMP greeting message invalid")

        self.monitor_in.write(json.dumps({"execute": "qmp_capabilities"}).encode("utf-8"))
        self.monitor_in.write("\n".encode("utf-8"))
        self.monitor_in.flush()

        answer = self._read_parse_json()
        if not "return" in answer:
            raise QMPError(f"Could not connect to QMP: {answer}")

    def _read_parse_json(self):
        line = self.monitor_out.readline().decode("utf-8")
        self.logger.debug("Received line: %s", line.rstrip("\r\n"))
        if not line:
            raise QMPError("Received empty response")
        return json.loads(line)

    def execute(self, command, arguments={}):
        json_command = {"execute": command, "arguments": arguments}

        self.monitor_in.write(json.dumps(json_command).encode("utf-8"))
        self.monitor_in.write("\n".encode("utf-8"))
        self.monitor_in.flush()

        answer = self._read_parse_json()
        # skip all asynchronous events
        while answer.get("event"):
            answer = self._read_parse_json()
        if "error" in answer:
            raise QMPError(answer["error"])
        return answer["return"]


@attr.s(eq=False)
class QMPError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))



import pytest

from labgrid.driver.manualswitchdriver import ManualSwitchDriver


class TestManualSwitchDriver:
    def test_create(self, target):
        d = ManualSwitchDriver(target, "foo-switch")
        assert isinstance(d, ManualSwitchDriver)

    def test_set_on(self, target, mocker):
        m = mocker.patch("builtins.input")

        d = ManualSwitchDriver(target, "foo-switch")
        target.activate(d)
        d.set(True)

        m.assert_called_once_with("Set foo-switch for target Test to ON and press enter")

    def test_set_off(self, target, mocker):
        m = mocker.patch("builtins.input")

        d = ManualSwitchDriver(target, "foo-switch")
        target.activate(d)
        d.set(False)

        m.assert_called_once_with("Set foo-switch for target Test to OFF and press enter")

    def test_get(self, target, mocker):
        m = mocker.patch("builtins.input")

        d = ManualSwitchDriver(target, "foo-switch")
        target.activate(d)

        d.set(True)
        assert d.get() is True

        d.set(False)
        assert d.get() is False



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class HTTPVideoStream(Resource):
    url = attr.ib(validator=attr.validators.instance_of(str))



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class HttpDigitalOutput(Resource):
    """This resource describes a generic HTTP-controlled output pin.

    Args:
        url (str): URL to use for setting a new state
        body_asserted (str): Request body to send to assert the output
        body_deasserted (str): Request body to send to de-assert the output
        method (str): HTTP method to use instead of PUT (the default) to set a new state

        url_get (str): URL to use for getting the state
        body_get_asserted (str): Regular Expression that matches an asserted response body
        body_get_deasserted (str): Regular Expression that matches a de-asserted response body
    """

    url = attr.ib(validator=attr.validators.instance_of(str))
    body_asserted = attr.ib(validator=attr.validators.instance_of(str))
    body_deasserted = attr.ib(validator=attr.validators.instance_of(str))
    method = attr.ib(default="PUT", validator=attr.validators.instance_of(str))

    url_get = attr.ib(default="", validator=attr.validators.instance_of(str))
    body_get_asserted = attr.ib(default="", validator=attr.validators.instance_of(str))
    body_get_deasserted = attr.ib(default="", validator=attr.validators.instance_of(str))



import enum

import attr

from labgrid import target_factory, step
from labgrid.driver import QuartusHPSDriver, SerialDriver
from labgrid.protocol import PowerProtocol
from labgrid.strategy import Strategy, never_retry


@attr.s(eq=False)
class StrategyError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))


class Status(enum.Enum):
    unknown = 0
    flashed_xload = 1
    flashed = 2


@target_factory.reg_driver
@attr.s(eq=False)
class QuartusHPSStrategy(Strategy):
    """QuartusHPSStrategy - Strategy to flash QSPI via 'Quartus Prime Programmer and Tools'"""

    bindings = {
        "power": PowerProtocol,
        "quartushps": QuartusHPSDriver,
        "serial": SerialDriver,
    }

    image = attr.ib(validator=attr.validators.instance_of(str))
    image_xload = attr.ib(validator=attr.validators.instance_of(str))
    status = attr.ib(default=Status.unknown)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @never_retry
    @step(args=["status"])
    def transition(self, status, *, step):
        if not isinstance(status, Status):
            status = Status[status]
        if status == Status.unknown:
            raise StrategyError(f"can not transition to {status}")
        elif status == self.status:
            step.skip("nothing to do")
            return  # nothing to do
        elif status == Status.flashed_xload:
            self.target.activate(self.power)
            self.power.cycle()
            self.target.activate(self.quartushps)
            # flash bootloader xload image to 0x0
            self.quartushps.flash(self.image_xload, 0x0)
        elif status == Status.flashed:
            self.transition(Status.flashed_xload)
            # flash bootloader image to 0x40000
            self.quartushps.flash(self.image, 0x40000)
            self.power.cycle()
            # activate serial in order to make 'labgrid-client -s $STATE con' work
            self.target.activate(self.serial)
        else:
            raise StrategyError(f"no transition found from {self.status} to {status}")
        self.status = status



import atexit
import tempfile
import logging
import shutil
import subprocess
import os
from select import select
from functools import wraps
from typing import Dict

import attr
from ..driver.exception import ExecutionError

from .helper import get_free_port, processwrapper

__all__ = ['sshmanager', 'SSHConnection', 'ForwardError']

def get_ssh_connect_timeout():
    return int(os.environ.get("LG_SSH_CONNECT_TIMEOUT", 30))


@attr.s
class SSHConnectionManager:
    """The SSHConnectionManager manages multiple SSH connections. This class
    should not be directly instantiated, use the exported sshmanager from this
    module instead.
    """
    _connections: 'Dict[str, SSHConnection]' = attr.ib(
        default=attr.Factory(dict),
        init=False,
        validator=attr.validators.optional(attr.validators.instance_of(dict))
    )

    def __attrs_post_init__(self):
        self.logger = logging.getLogger(f"{self}")
        atexit.register(self.close_all)

    def get(self, host: str):
        """Retrieve or create a new connection to a given host

        Arguments:
            host (str): host to retrieve the connection for

        Returns:
            :obj:`SSHConnection`: the SSHConnection for the host"""
        instance = self._connections.get(host)
        if instance is None:
            self.logger.debug("Creating SSHConnection for %s", host)
            instance = SSHConnection(host)
            instance.connect()
            self._connections[host] = instance
        return instance

    def add_connection(self, connection):
        """Add an existing SSHConnection to the manager.
        This is useful for manually created connections which should also be
        available via the manager.

        Arguments:
            connection (:obj:`SSHConnection`): SSHconnection to add to the manager
        """
        assert isinstance(connection, SSHConnection)
        if connection.host not in self._connections:
            self._connections[connection.host] = connection

    def remove_connection(self, connection):
        assert isinstance(connection, SSHConnection)
        if connection.isconnected():
            raise ExecutionError("Can't remove connected connection")
        del self._connections[connection.host]

    def remove_by_name(self, name):
        del self._connections[name]

    def open(self, host):
        return self.get(host)

    def close(self, host):
        con = self.get(host)
        con.disconnect()
        self.remove_connection(con)

    def request_forward(self, host, dest, port):
        con = self.get(host)
        return con.add_port_forward(dest, port)

    def remove_forward(self, host, dest, port):
        con = self.get(host)
        con.remove_port_forward(dest, port)

    def put_file(self, host, local_file, remote_file):
        con = self.get(host)
        con.put_file(local_file, remote_file)

    def get_file(self, host, remote_file, local_file):
        con = self.get(host)
        con.get_file(remote_file, local_file)

    def close_all(self):
        """Close all open connections and remove them from the manager """
        for name, connection in self._connections.items():
            if connection.isconnected():
                connection.disconnect()
        names = self._connections.copy().keys()
        for name in names:
            self.remove_by_name(name)


def _check_connected(func):
    """Check if an SSHConnection is connected as a decorator"""

    @wraps(func)
    def wrapper(self, *_args, **_kwargs):
        if not self.isconnected():
            raise ExecutionError(
                f"{func.__qualname__} can not be called ({self} is not connected)"
            )
        return func(self, *_args, **_kwargs)

    return wrapper


@attr.s
class SSHConnection:
    """SSHConnections are individual connections to hosts managed by a control
    socket. In addition to command execution this class also provides an
    interface to manage port forwardings. These are used in the remote
    infrastructure to tunnel multiple connections over one SSH link.

    A public identity infrastructure is assumed, no extra username or passwords
    are supported."""
    host = attr.ib(validator=attr.validators.instance_of(str))
    _connected = attr.ib(
        default=False, init=False, validator=attr.validators.instance_of(bool)
    )
    _tmpdir = attr.ib(
        default=attr.Factory(lambda: tempfile.mkdtemp(prefix="lg-con-")),
        init=False,
        validator=attr.validators.instance_of(str)
    )
    _l_forwards = attr.ib(init=False, default=attr.Factory(dict))
    _r_forwards = attr.ib(init=False, default=attr.Factory(set))

    def __attrs_post_init__(self):
        self._logger = logging.getLogger(f"{self}")
        self._socket = None
        self._master = None
        self._keepalive = None
        atexit.register(self.cleanup)

    @staticmethod
    def _get_ssh_base_args():
        return ["-x", "-o", "LogLevel=ERROR", "-o", "PasswordAuthentication=no"]

    def _get_ssh_control_args(self):
        if self._socket:
            return [
                "-o", "ControlMaster=no",
                "-o", f"ControlPath={self._socket}",
            ]
        return []

    def _get_ssh_args(self):
        args = SSHConnection._get_ssh_base_args()
        args += self._get_ssh_control_args()
        return args

    def _open_connection(self):
        """Internal function which appends the control socket and checks if the
        connection is already open"""
        if self._check_external_master():
            self._logger.info("Using existing SSH connection to %s", self.host)
        else:
            self._start_own_master()
            self._logger.info("Created new SSH connection to %s", self.host)
        self._start_keepalive()
        self._connected = True

    def _run_socket_command(self, command, forward=None):
        "Internal function to send a command to the control socket"
        complete_cmd = ["ssh"] + self._get_ssh_args()
        complete_cmd += ["-O", command]
        if forward:
            for item in forward:
                complete_cmd.append(item)
        complete_cmd.append(self.host)
        self._logger.debug("Running control command: %s", " ".join(complete_cmd))
        subprocess.check_call(
            complete_cmd,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            timeout=2,
        )

    @_check_connected
    def get_prefix(self):
        return ["ssh"] + self._get_ssh_args() + [self.host]

    @_check_connected
    def run(self, command, *, codec="utf-8", decodeerrors="strict",
            force_tty=False, stderr_merge=False, stderr_loglevel=None,
            stdout_loglevel=None):

        """Run a command over the SSHConnection

        Args:
            command (string): The command to run
            codec (string, optional): output encoding. Defaults to "utf-8".
            decodeerrors (string, optional): behavior on decode errors. Defaults
                 to "strict". Refer to stdtypes' bytes.decode for details.
            force_tty (bool, optional): force allocate a tty (ssh -tt). Defaults
                 to False
            stderr_merge (bool, optional): merge ssh subprocess stderr into
                 stdout. Defaults to False.
            stdout_loglevel (int, optional): log stdout with specific log level
                 as well. Defaults to None, i.e. don't log.
            stderr_loglevel (int, optional): log stderr with specific log level
                 as well. Defaults to None, i.e. don't log.

        returns:
            (stdout, stderr, returncode)
        """

        complete_cmd = ["ssh"] + self._get_ssh_args()
        if force_tty:
            complete_cmd += ["-tt"]
        complete_cmd += [self.host, command]
        self._logger.debug("Sending command: %s", " ".join(complete_cmd))
        if stderr_merge:
            stderr_pipe = subprocess.STDOUT
        else:
            stderr_pipe = subprocess.PIPE
        try:
            sub = subprocess.Popen(
                complete_cmd, stdout=subprocess.PIPE, stderr=stderr_pipe,
                stdin=subprocess.DEVNULL
            )
        except:
            raise ExecutionError(
                f"error executing command: {complete_cmd}"
            )

        stdout = []
        stderr = []

        readable = {
            sub.stdout.fileno(): (sub.stdout, stdout, stdout_loglevel),
        }

        if sub.stderr is not None:
            readable[sub.stderr.fileno()] = (sub.stderr, stderr, stderr_loglevel)

        while readable:
            for fd in select(readable, [], [])[0]:
                stream, output, loglevel = readable[fd]
                line = stream.readline().decode(codec, decodeerrors)
                if line == "": # EOF
                    del readable[fd]
                else:
                    line = line.rstrip('\n')
                    output.append(line)
                    if loglevel is not None:
                        self._logger.log(loglevel, line)

        sub.communicate()
        return stdout, stderr, sub.returncode

    def run_check(self, command, *, codec="utf-8", decodeerrors="strict",
                  force_tty=False, stderr_merge=False, stderr_loglevel=None,
                  stdout_loglevel=None):
        """
        Runs a command over the SSHConnection
        returns the output if successful, raises ExecutionError otherwise.

        Except for the means of returning the value, this is equivalent to
        run.

        Args:
            command (string): The command to run
            codec (string, optional): output encoding. Defaults to "utf-8".
            decodeerrors (string, optional): behavior on decode errors. Defaults
                 to "strict". Refer to stdtypes' bytes.decode for details.
            force_tty (bool, optional): force allocate a tty (ssh -tt). Defaults
                 to False
            stderr_merge (bool, optional): merge ssh subprocess stderr into
                 stdout. Defaults to False.
            stdout_loglevel (int, optional): log stdout with specific log level
                 as well. Defaults to None, i.e. don't log.
            stderr_loglevel (int, optional): log stderr with specific log level
                 as well. Defaults to None, i.e. don't log.

        Returns:
            List[str]: stdout of the executed command if successful and
                       otherwise an ExecutionError Exception

        """
        stdout, stderr, exitcode = self.run(
            command, codec=codec, decodeerrors=decodeerrors, force_tty=force_tty,
            stderr_merge=stderr_merge, stdout_loglevel=stdout_loglevel,
            stderr_loglevel=stderr_loglevel
        )

        if exitcode != 0:
            raise ExecutionError(command, stdout, stderr)
        return stdout

    @_check_connected
    def get_file(self, remote_file, local_file):
        """Get a file from the remote host"""
        complete_cmd = ["scp"] + self._get_ssh_control_args()
        complete_cmd += [
            f"{self.host}:{remote_file}",
            f"{local_file}"
        ]
        self._logger.debug("Running command: %s", complete_cmd)
        subprocess.check_call(
            complete_cmd,
            stdin=subprocess.DEVNULL,
        )

    @_check_connected
    def put_file(self, local_file, remote_path):
        """Put a file onto the remote host"""
        complete_cmd = ["rsync", "--compress", "--sparse", "--copy-links", "--verbose", "--progress", "--times", "-e",
                        " ".join(['ssh'] + self._get_ssh_args())]
        complete_cmd += [
            f"{local_file}",
            f"{self.host}:{remote_path}"
        ]
        self._logger.debug("Running command: %s", complete_cmd)
        processwrapper.check_output(
            complete_cmd,
            stdin=subprocess.DEVNULL,
            print_on_silent_log=True
        )

    @_check_connected
    def add_port_forward(self, remote_host, remote_port, local_port=None):
        """forward command"""
        if local_port is None:
            local_port = get_free_port()
        destination = f"{remote_host}:{remote_port}"

        if destination in self._l_forwards:
            return self._l_forwards[destination]
        self._run_socket_command(
            "forward", [
                f"-L{local_port}:{destination}"
            ]
        )
        self._l_forwards[destination] = local_port
        return local_port

    @_check_connected
    def remove_port_forward(self, remote_host, remote_port):
        """cancel command"""
        destination = f"{remote_host}:{remote_port}"
        local_port = self._l_forwards.pop(destination, None)

        if local_port is None:
            raise ForwardError("Forward does not exist")

        self._run_socket_command(
            "cancel", [
                f"-L{local_port}:{destination}"
            ]
        )

    @_check_connected
    def add_remote_port_forward(self, remote_port, local_port, remote_bind=None):
        """remote forward command

        Note that the remote socket is not *bound* to any specific IP by
        default, making it reachable by the target. Also, 'GatewayPorts
        clientspecified' needs to be configured in the remote host's
        sshd_config.
        """
        if remote_bind is None:
            remote_bind = "*"

        forward = f"-R{remote_bind}:{remote_port:d}:localhost:{local_port:d}"

        self._run_socket_command("forward", [forward])
        self._r_forwards.add(forward)

    @_check_connected
    def remove_remote_port_forward(self, remote_port, local_port, remote_bind=None):
        """remote cancel command"""
        if remote_bind is None:
            remote_bind = "*"

        forward = f"-R{remote_bind}:{remote_port:d}:localhost:{local_port:d}"

        self._r_forwards.remove(forward)
        self._run_socket_command("cancel", [forward])

    def connect(self):
        if not self._connected:
            self._open_connection()

    def isconnected(self):
        return self._connected and self._check_keepalive()

    def _check_external_master(self):
        args = ["ssh", "-O", "check", f"{self.host}"]
        # We don't want to confuse the use with SSHs output here, so we need to
        # capture and parse it.
        proc = subprocess.Popen(
            args,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        stdout, _ = proc.communicate(timeout=60)
        check = proc.wait()
        if check == 0:
            self._logger.debug("Found existing control socket")
            return True
        if b"No such file or directory" in stdout or b"No ControlPath specified" in stdout:
            self._logger.debug("No existing control socket found")
            return False

        self._logger.debug("Unexpected ssh check output '%s'", stdout)
        return False

    def _start_own_master(self):
        """Starts a controlmaster connection in a temporary directory."""
        control = os.path.join(self._tmpdir, f'control-{self.host}')

        connect_timeout = get_ssh_connect_timeout()

        self._logger.debug("ControlSocket: %s", control)
        args = ["ssh"] + SSHConnection._get_ssh_base_args()
        args += [
            "-n", "-MN",
            "-o", f"ConnectTimeout={connect_timeout}",
            "-o", "ControlPersist=300",
            "-o", "ControlMaster=yes",
            "-o", f"ControlPath={control}",
            # We don't want to ask the user to confirm host keys here.
            "-o", "StrictHostKeyChecking=yes",
            self.host,
        ]

        self._logger.debug("Master Start command: %s", " ".join(args))
        assert self._master is None
        self._master = subprocess.Popen(
            args,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        try:
            stdout, stderr = self._master.communicate(timeout=connect_timeout)
            if self._master.returncode != 0:
                raise ExecutionError(
                    f"failed to connect to {self.host} with args {args}, returncode={self._master.returncode} {stdout},{stderr}"  # pylint: disable=line-too-long
                )
        except subprocess.TimeoutExpired:
            self._master.kill()
            stdout, stderr = self._master.communicate()
            raise ExecutionError(
                f"failed to connect (timeout) to {self.host} with args {args}, process killed, got {stdout},{stderr}"  # pylint: disable=line-too-long
            )

        if not os.path.exists(control):
            raise ExecutionError(f"no control socket to {self.host}")

        self._socket = control

        self._logger.debug('Connected to %s', self.host)

    def _stop_own_master(self):
        assert self._socket is not None
        assert self._master is not None

        try:
            self._run_socket_command("cancel")
            self._run_socket_command("exit")
            # if the master doesn't terminate in less than 60 seconds,
            # something is very wrong
            self._master.communicate(timeout=60)
        finally:
            self._socket = None
            self._master = None

    def _start_keepalive(self):
        """Starts a keepalive connection via the own or external master."""
        args = ["ssh"] + self._get_ssh_args() + [self.host, "cat"]

        assert self._keepalive is None
        self._keepalive = subprocess.Popen(
            args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            start_new_session=True,
        )

        self._logger.debug('Started keepalive for %s', self.host)

    def _check_keepalive(self):
        return self._keepalive.poll() is None

    def _stop_keepalive(self):
        assert self._keepalive is not None

        self._logger.debug('Stopping keepalive for %s', self.host)

        try:
            self._keepalive.communicate(timeout=60)
        except subprocess.TimeoutExpired:
            self._keepalive.kill()
            self._keepalive.communicate(timeout=60)
        finally:
            self._keepalive = None

    @_check_connected
    def disconnect(self):
        assert self._connected
        try:
            self._stop_keepalive()

            if self._socket:
                self._logger.info("Closing SSH connection to %s", self.host)
                self._stop_own_master()
        finally:
            self._connected = False

    def cleanup(self):
        if self.isconnected():
            # cancel local forwards
            for destination, local_port in self._l_forwards.items():
                self._run_socket_command("cancel", [f"-L{local_port}:{destination}"])
            self._l_forwards.clear()
            # cancel remote forwards
            for forward in self._r_forwards:
                self._run_socket_command("cancel", [forward])
            self._r_forwards.clear()
            self.disconnect()
        shutil.rmtree(self._tmpdir)

sshmanager = SSHConnectionManager()

@attr.s
class ForwardError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))



import attr

from .common import Driver
from ..factory import target_factory
from ..resource.remote import NetworkDeditecRelais8
from ..step import step
from ..protocol import DigitalOutputProtocol
from ..util.agentwrapper import AgentWrapper


@target_factory.reg_driver
@attr.s(eq=False)
class DeditecRelaisDriver(Driver, DigitalOutputProtocol):
    bindings = {
        "relais": {"DeditecRelais8", NetworkDeditecRelais8},
    }

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.wrapper = None

    def on_activate(self):
        if isinstance(self.relais, NetworkDeditecRelais8):
            host = self.relais.host
        else:
            host = None
        self.wrapper = AgentWrapper(host)
        self.proxy = self.wrapper.load("deditec_relais8")

    def on_deactivate(self):
        self.wrapper.close()
        self.wrapper = None
        self.proxy = None

    @Driver.check_active
    @step(args=["status"])
    def set(self, status):
        if self.relais.invert:
            status = not status
        self.proxy.set(self.relais.busnum, self.relais.devnum, self.relais.index, status)

    @Driver.check_active
    @step(result=True)
    def get(self):
        status = self.proxy.get(self.relais.busnum, self.relais.devnum, self.relais.index)
        if self.relais.invert:
            status = not status
        return status



import enum
import os
import pathlib
import time
import subprocess

import attr

from ..factory import target_factory
from ..resource.remote import RemoteUSBResource
from ..step import step
from ..util.managedfile import ManagedFile
from .common import Driver
from ..driver.exception import ExecutionError

from ..util.helper import processwrapper
from ..util.agentwrapper import AgentWrapper
from ..util import Timeout


class Mode(enum.Enum):
    DD = "dd"
    BMAPTOOL = "bmaptool"

    def __str__(self):
        return self.value


@target_factory.reg_driver
@attr.s(eq=False)
class USBStorageDriver(Driver):
    bindings = {
        "storage": {
            "USBMassStorage",
            "NetworkUSBMassStorage",
            "USBSDMuxDevice",
            "NetworkUSBSDMuxDevice",
            "USBSDWireDevice",
            "NetworkUSBSDWireDevice",
            "USBSDWire3Device",
            "NetworkUSBSDWire3Device",
        },
    }
    image = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )
    WAIT_FOR_MEDIUM_TIMEOUT = 10.0 # s
    WAIT_FOR_MEDIUM_SLEEP = 0.5 # s
    MOUNT_RETRIES = 5

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.wrapper = None
        self.proxy = None

    def _start_wrapper(self):
        if self.wrapper:
            return
        host = self.storage.host if isinstance(self.storage, RemoteUSBResource) else None
        self.wrapper = AgentWrapper(host)
        self.proxy = self.wrapper.load('udisks2')

    def on_activate(self):
        pass

    def on_deactivate(self):
        if self.wrapper:
            self.wrapper.close()
        self.wrapper = None
        self.proxy = None

    @Driver.check_active
    @step(args=['sources', 'target', 'partition', 'target_is_directory'])
    def write_files(self, sources, target, partition, target_is_directory=True):
        """
        Write the file(s) specified by filename(s) to the
        bound USB storage partition.

        Args:
            sources (List[str]): path(s) to the file(s) to be copied to the bound USB storage
                partition.
            target (PurePath): target directory or file to copy to
            partition (int): mount the specified partition or None to mount the whole disk
            target_is_directory (bool): Whether target is a directory

        Raises:
            Exception if anything goes wrong
        """
        self._wait_for_medium(partition)

        self._start_wrapper()

        self.devpath = self._get_devpath(partition)
        mount_path = self.proxy.mount(self.devpath, self.MOUNT_RETRIES)

        try:
            # (pathlib.PurePath(...) / "/") == "/", so we turn absolute paths into relative
            # paths with respect to the mount point here
            target_rel = target.relative_to(target.root) if target.root is not None else target
            target_path = str(pathlib.PurePath(mount_path) / target_rel)

            copied_sources = []

            for f in sources:
                mf = ManagedFile(f, self.storage)
                mf.sync_to_resource()
                copied_sources.append(mf.get_remote_path())

            if target_is_directory:
                args = ["cp", "-t", target_path] + copied_sources
            else:
                if len(sources) != 1:
                    raise ValueError("single source argument required when target_is_directory=False")

                args = ["cp", "-T", copied_sources[0], target_path]

            processwrapper.check_output(self.storage.wrap_command(args))
            self.proxy.unmount(self.devpath)
        except:
            # We are going to die with an exception anyway, so no point in waiting
            # to make sure everything has been written before continuing
            self.proxy.unmount(self.devpath, lazy=True)
            raise

    @Driver.check_active
    @step(args=['filename'])
    def write_image(self, filename=None, mode=Mode.DD, partition=None, skip=0, seek=0):
        """
        Writes the file specified by filename or if not specified by config image subkey to the
        bound USB storage root device or partition.

        Args:
            filename (str): optional, path to the image to write to bound USB storage
            mode (Mode): optional, Mode.DD or Mode.BMAPTOOL (defaults to Mode.DD)
            partition (int or None): optional, write to the specified partition or None for writing
                to root device (defaults to None)
            skip (int): optional, skip n 512-sized blocks at start of input file (defaults to 0)
            seek (int): optional, skip n 512-sized blocks at start of output (defaults to 0)
        """
        if filename is None and self.image is not None:
            filename = self.target.env.config.get_image_path(self.image)
        assert filename, "write_image requires a filename"
        mf = ManagedFile(filename, self.storage)
        mf.sync_to_resource()

        self._wait_for_medium(partition)

        target = self._get_devpath(partition)
        remote_path = mf.get_remote_path()

        if mode == Mode.DD:
            self.logger.info('Writing %s to %s using dd.', remote_path, target)
            block_size = '512' if skip or seek else '4M'
            args = [
                "dd",
                f"if={remote_path}",
                f"of={target}",
                "oflag=direct",
                "status=progress",
                f"bs={block_size}",
                f"skip={skip}",
                f"seek={seek}",
                "conv=fdatasync"
            ]
        elif mode == Mode.BMAPTOOL:
            if skip or seek:
                raise ExecutionError("bmaptool does not support skip or seek")

            # Try to find a block map file using the same logic that bmaptool
            # uses. Handles cases where the image is named like: <image>.bz2
            # and the block map file is <image>.bmap
            mf_bmap = None
            image_path = filename
            while True:
                bmap_path = f"{image_path}.bmap"
                if os.path.exists(bmap_path):
                    mf_bmap = ManagedFile(bmap_path, self.storage)
                    mf_bmap.sync_to_resource()
                    break

                image_path, ext = os.path.splitext(image_path)
                if not ext:
                    break

            self.logger.info('Writing %s to %s using bmaptool.', remote_path, target)
            args = [
                "bmaptool",
                "copy",
                f"{remote_path}",
                f"{target}",
            ]

            if mf_bmap is None:
                args.append("--nobmap")
            else:
                args.append(f"--bmap={mf_bmap.get_remote_path()}")
        else:
            raise ValueError

        processwrapper.check_output(
            self.storage.command_prefix + args,
            print_on_silent_log=True
        )

    def _get_devpath(self, partition):
        partition = "" if partition is None else partition
        # simple concatenation is sufficient for USB mass storage
        return f"{self.storage.path}{partition}"

    @Driver.check_active
    def _wait_for_medium(self, partition):
        timeout = Timeout(self.WAIT_FOR_MEDIUM_TIMEOUT)
        while not timeout.expired:
            if self.get_size(partition) > 0:
                break
            time.sleep(self.WAIT_FOR_MEDIUM_SLEEP)
        else:
            raise ExecutionError("Timeout while waiting for medium")

    @Driver.check_active
    @step(args=['partition'], result=True)
    def get_size(self, partition=None):
        """
        Get the size of the bound USB storage root device or partition.

        Args:
            partition (int or None): optional, get size of the specified partition or None for
                getting the size of the root device (defaults to None)

        Returns:
            int: size in bytes, or 0 if the partition is not found
        """
        args = ["cat", f"/sys/class/block/{self._get_devpath(partition)[5:]}/size"]
        try:
            size = subprocess.check_output(self.storage.wrap_command(args))
        except subprocess.CalledProcessError:
            # while the medium is getting ready, the file does not yet exist
            return 0
        try:
            return int(size) * 512
        except ValueError:
            # when the medium gets ready the sysfs attribute is empty for a short time span
            return 0



import enum

import attr

from labgrid.driver import BareboxDriver, ShellDriver, USBSDMuxDriver
from labgrid import step, target_factory
from labgrid.protocol import PowerProtocol
from labgrid.strategy import Strategy, never_retry


@attr.s(eq=False)
class StrategyError(Exception):
    msg = attr.ib(validator=attr.validators.instance_of(str))


class Status(enum.Enum):
    unknown = 0
    barebox = 1
    shell = 2


@target_factory.reg_driver
@attr.s(eq=False)
class ExampleStrategy(Strategy):
    """ExampleStrategy - Strategy to for the usbpower labgrid example"""

    bindings = {
        "power": PowerProtocol,
        "sdmux": USBSDMuxDriver,
        "barebox": BareboxDriver,
        "shell": ShellDriver,
    }

    status = attr.ib(default=Status.unknown)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @never_retry
    @step(args=["status"])
    def transition(self, status, *, step):
        if not isinstance(status, Status):
            status = Status[status]
        if status == Status.unknown:
            raise StrategyError(f"can not transition to {status}")
        elif status == self.status:
            step.skip("nothing to do")
            return  # nothing to do
        elif status == Status.barebox:
            self.target.activate(self.power)
            self.target.activate(self.sdmux)
            # power off
            self.power.off()
            # configure sd-mux
            self.sdmux.set_mode("dut")
            # cycle power
            self.power.on()
            # interrupt barebox
            self.target.activate(self.barebox)
        elif status == Status.shell:
            # transition to barebox
            self.transition(Status.barebox)
            self.barebox.boot("")
            self.barebox.await_boot()
            self.target.activate(self.shell)
        else:
            raise StrategyError(f"no transition found from {self.status} to {status}")
        self.status = status



from labgrid.remote.common import Place, ResourceMatch, Reservation, set_map_from_dict, build_dict_from_map
import labgrid.remote.generated.labgrid_coordinator_pb2 as labgrid_coordinator_pb2

def test_place_as_pb2():
    place = Place(name="testing-place")
    pb2 = place.as_pb2()
    assert pb2.name == "testing-place"
    assert pb2.created == place.created
    assert pb2.changed == place.changed

def test_place_from_pb2():
    place_start = Place(name="testing-place", comment="such-comment")
    pb2 = place_start.as_pb2()
    assert pb2.name == "testing-place"
    place_new = Place.from_pb2(pb2)
    assert place_new.name == "testing-place"
    assert place_new.name == place_start.name
    assert place_new.comment == place_start.comment
    assert place_new.tags == place_start.tags
    assert place_new.matches == place_start.matches
    assert place_new.acquired == place_start.acquired
    assert place_new.acquired_resources == place_start.acquired_resources
    assert place_new.allowed == place_start.allowed
    assert place_new.created == place_start.created
    assert place_new.changed == place_start.changed
    assert place_new.reservation == place_start.reservation

def test_from_pb2_tags():
    tags = {"some": "test", "more": "values"}
    place_start = Place(name="testing-place", tags=tags)
    pb2 = place_start.as_pb2()
    assert pb2.name == "testing-place", f"PB2 has wrong name: {pb2}"
    assert pb2.tags is not None, f"PB2 has no tags field: {pb2}"
    place_new = Place.from_pb2(pb2)
    assert place_new.name == "testing-place"
    assert place_new.tags == place_start.tags
    assert place_new.tags == tags

def test_from_pb2_matches():
    rm = ResourceMatch("such", "test", "match")
    place_start = Place(name="testing-place", matches=[rm])
    pb2 = place_start.as_pb2()
    assert pb2.name == "testing-place", f"PB2 has wrong name: {pb2}"
    assert pb2.tags is not None, f"PB2 has no tags field: {pb2}"
    place_new = Place.from_pb2(pb2)
    assert place_new.name == "testing-place"
    assert place_new.tags == place_start.tags
    assert place_new.matches == place_start.matches

def test_from_pb2_tags_deepcopy():
    # Used by the RemotePlaceManager
    tags = {"some": "test", "more": "values"}
    place_start = Place(name="testing-place", tags=tags)
    pb2 = place_start.as_pb2()
    place_new = Place.from_pb2(pb2)
    import copy
    tags_copy = copy.deepcopy(place_new.tags)

def test_place_as_pb2_copy_with_match():
    tags = {"some": "test", "more": "values"}
    # Used by the RemotePlaceManager
    place_start = Place(name="testing-place", tags=tags, comment="Hello", aliases={"some": "alias"}, matches=[ResourceMatch("testporter","somegroup","someclass")])
    out = labgrid_coordinator_pb2.ClientOutMessage()

    update_response = labgrid_coordinator_pb2.UpdateResponse()
    update_response.place.CopyFrom(place_start.as_pb2())

    out.updates.append(update_response)

def test_match_as_from_pb2():
    rms = ResourceMatch("*", "somegroup", "someclass")
    pb2 = rms.as_pb2()
    assert pb2
    rme = ResourceMatch.from_pb2(pb2)
    assert rms == rme

def test_reservation_as_pb2():
    reservation = Reservation(
        owner="test",
        filters={
            "main": {"some": "filter"},
        },
        allocations={
            "main": ["the-place"],
        },
    )
    pb2 = reservation.as_pb2()
    assert pb2.owner == "test"
    assert pb2.token == reservation.token
    assert pb2.state == reservation.state.value
    assert pb2.filters["main"].filter == {"some": "filter"}
    assert pb2.created == reservation.created
    assert pb2.timeout == reservation.timeout

def test_reservation_as_from_pb2():
    resold = Reservation(
        owner="test",
        filters={
            "main": {"some": "filter"},
        },
        allocations={
            "main": ["the-place"],
        },
    )
    pb2 = resold.as_pb2()
    assert pb2.owner == resold.owner
    assert pb2.token == resold.token
    assert pb2.state == resold.state.value
    assert pb2.filters["main"].filter == {"some": "filter"}
    assert pb2.created == resold.created
    assert pb2.timeout == resold.timeout

    resnew = Reservation.from_pb2(pb2)

    assert resnew.owner == resold.owner
    assert resnew.token == resold.token
    assert resnew.state == resold.state
    assert resnew.filters["main"] == resold.filters["main"]
    assert resnew.created == resold.created
    assert resnew.timeout == resold.timeout

def test_resource_dict():
    params = {
        'host': 'foo',
        'model_id': 2277,
        'vendor_id': 1133,
        'path': None,
    }

    resource = labgrid_coordinator_pb2.Resource()
    set_map_from_dict(resource.params, params)
    bm = resource.SerializeToString()

    resource = labgrid_coordinator_pb2.Resource()
    resource.ParseFromString(bm)
    decoded = build_dict_from_map(resource.params)

    assert params == decoded

def test_map_serialize(benchmark):
    params = {
        'host': 'foo',
        'model_id': 2277,
        'vendor_id': 1133,
        'path': None,
    }

    def run():
        resource = labgrid_coordinator_pb2.Resource()
        set_map_from_dict(resource.params, params)
        bm = resource.SerializeToString()

    benchmark(run)

def test_map_deser(benchmark):
    params = {
        'host': 'foo',
        'model_id': 2277,
        'vendor_id': 1133,
        'path': None,
    }

    resource = labgrid_coordinator_pb2.Resource()
    set_map_from_dict(resource.params, params)
    bm = resource.SerializeToString()

    def run():
        resource = labgrid_coordinator_pb2.Resource()
        resource.ParseFromString(bm)
        decoded = build_dict_from_map(resource.params)

    benchmark(run)



from time import sleep

import pytest

from labgrid import step, steps


@step()
def step_a(*, step):
    assert steps.get_current() is not None
    return step.level


@step()
def step_outer(*, step):
    assert step.level == 1
    return step_a()


def test_single():
    assert steps.get_current() is None
    step_a()
    assert steps.get_current() is None


def test_nested():
    assert steps.get_current() is None
    inner_level = step_outer()
    assert steps.get_current() is None
    assert inner_level == 2


@step()
def step_sleep(*, step):
    sleep(0.25)
    return step


def test_timing():
    step = step_sleep()
    assert step.duration == pytest.approx(0.25, abs=1e-2)
    assert step.exception is None


class A:
    @step()
    def method_step(self, foo, *, step):
        return step

    @step(args=["foo"])
    def method_args_step(self, foo, *, step):
        return step

    @step(title="test-title")
    def method_title_step(self, foo, *, step):
        return step

    @step(result=True)
    def method_result_step(self, foo, *, step):
        return step

    @step(tag="dummy")
    def method_tag_step(self, foo, *, step):
        return step


def test_method():
    a = A()
    step = a.method_step("foo")
    assert step.source == a
    assert step.title == "method_step"
    assert step.args is None
    assert step.result is None
    assert step.tag is None


def test_method_args():
    a = A()
    step = a.method_args_step("foo")
    assert step.source == a
    assert step.title == "method_args_step"
    assert step.args == {"foo": "foo"}
    assert step.result is None
    assert step.tag is None


def test_method_title():
    a = A()
    step = a.method_title_step("foo")
    assert step.source == a
    assert step.title == "test-title"
    assert step.args is None
    assert step.result is None
    assert step.tag is None


def test_method_result():
    a = A()
    step = a.method_result_step("foo")
    assert step.source == a
    assert step.title == "method_result_step"
    assert step.args is None
    assert step.result is step
    assert step.tag is None


def test_method_tag():
    a = A()
    step = a.method_tag_step("foo")
    assert step.source == a
    assert step.title == "method_tag_step"
    assert step.args is None
    assert step.result is None
    assert step.tag == "dummy"


@step(args=["default"])
def step_default_arg(default=None, *, step):
    return step


def test_default_arg():
    step = step_default_arg()
    assert step.args["default"] == None

    step = step_default_arg(default="real")
    assert step.args["default"] == "real"


@step()
def step_error(output, *, step):
    output.append(step)
    raise ValueError("dummy")


def test_error():
    output = []
    with pytest.raises(ValueError, match=r"dummy"):
        step_error(output)
    step = output[0]
    assert step.exception is not None
    assert isinstance(step.exception, ValueError)


@step()
def step_event_skip(*, step):
    step.skip("testing")


def test_event():
    events = []

    def callback(event):
        events.append(event)

    steps.subscribe(callback)
    try:
        step = step_event_skip()
    finally:
        steps.unsubscribe(callback)

    skip_event = [e for e in events if "skip" in e.data]
    assert len(skip_event) == 1
    assert skip_event[0].data["skip"] == "testing"


def test_subscriber_error():
    events = []

    def callback(event):
        raise ValueError("from callback")

    steps.subscribe(callback)
    with pytest.warns(UserWarning):
        step = step_event_skip()
    steps.unsubscribe(callback)



import attr

from ..factory import target_factory
from .common import Resource


@target_factory.reg_resource
@attr.s(eq=False)
class NetworkPowerPort(Resource):
    """The NetworkPowerPort describes a remotely switchable PowerPort

    Args:
        model (str): model of the external power switch
        host (str): host to connect to
        index (str): index of the power port on the external switch
    """
    model = attr.ib(validator=attr.validators.instance_of(str))
    host = attr.ib(validator=attr.validators.instance_of(str))
    index = attr.ib(validator=attr.validators.instance_of(str),
                    converter=lambda x: str(int(x)))


@target_factory.reg_resource
@attr.s(eq=False)
class PDUDaemonPort(Resource):
    """The PDUDaemonPort describes a port on a PDU accessible via PDUDaemon

    Args:
        host (str): name of the host running the PDUDaemon
        pdu (str): name of the PDU in the configuration file
        index (int): index of the power port on the PDU
    """
    host = attr.ib(validator=attr.validators.instance_of(str))
    pdu = attr.ib(validator=attr.validators.instance_of(str))
    index = attr.ib(validator=attr.validators.instance_of(int), converter=int)



import os
import pexpect
import pytest

@pytest.fixture
def short_env(tmpdir):
    p = tmpdir.join("config.yaml")
    p.write(
"""
targets:
  test1:
    drivers: {}
  test2:
    role: foo
    resources: {}
"""
    )
    return p

@pytest.fixture
def short_test(tmpdir):
    t = tmpdir.join("test.py")
    t.write(
"""
def test(env):
    assert True
"""
    )
    return t

def test_config(short_test):
    with pexpect.spawn(f'pytest --traceconfig {short_test}') as spawn:
        spawn.expect(pexpect.EOF)
        assert b'labgrid.pytestplugin' in spawn.before
        spawn.close()
        assert spawn.exitstatus == 0

def test_env_fixture(short_env, short_test):
    with pexpect.spawn(f'pytest --lg-env {short_env} {short_test}') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0

def test_env_fixture_no_logging(short_env, short_test):
    with pexpect.spawn(f'pytest -p no:logging --lg-env {short_env} {short_test}') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0, spawn.before

def test_env_env_fixture(short_env, short_test):
    env=os.environ.copy()
    env['LG_ENV'] = short_env
    with pexpect.spawn(f'pytest {short_test}') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0

def test_env_with_junit(short_env, short_test, tmpdir):
    x = tmpdir.join('junit.xml')
    with pexpect.spawn(f'pytest --junitxml={x} --lg-env {short_env} {short_test}') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0

def test_help(short_test, monkeypatch):
    # argparse in Python >= 3.14 enables colored output by default,
    # disable that to allow argument assertions below
    monkeypatch.setenv("NO_COLOR", "1")

    with pexpect.spawn(f'pytest --help {short_test}') as spawn:
        spawn.expect(pexpect.EOF)
        assert b'--lg-coordinator=COORDINATOR_ADDRESS' in spawn.before
        spawn.close()
        assert spawn.exitstatus == 0

def test_help_coordinator(short_test):
    with pexpect.spawn(f'pytest --lg-coordinator=127.0.0.1:20408 --help {short_test}') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        assert spawn.exitstatus == 0

def test_log_without_capturing(short_env, short_test, tmpdir):
    with pexpect.spawn(f'pytest -vv -s --lg-env {short_env} {short_test}') as spawn:
        spawn.expect(pexpect.EOF)
        spawn.close()
        print(spawn.before)
        assert spawn.exitstatus == 0
